"""
Utilities for unicode IO.
"""

import csv
import sys

from pkg_resources import resource_string

import yaml


__all__ = [
    'csv_unicode_reader',
    'yaml_load_unicode',
    'load_yaml_resource',
]


if sys.version_info[0] < 3:
    def csv_unicode_reader(lines, encoding='utf-8', **kwargs):
        """Load unicode CSV file."""
        return [[r.decode(encoding) for r in row]
                for row in csv.reader(stream, **kwargs)]

else:
    def csv_unicode_reader(lines, encoding='utf-8', **kwargs):
        """Load unicode CSV file."""
        lines = [l.decode(encoding) for l in lines]
        return csv.reader(lines, **kwargs)


def yaml_load_unicode(stream, Loader=yaml.SafeLoader):
    """Load YAML with all strings loaded as unicode objects."""
    class UnicodeLoader(Loader):
        pass
    def construct_yaml_str(self, node):
        # Override the default string handling function
        # to always return unicode objects
        return self.construct_scalar(node)
    UnicodeLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG,
        construct_yaml_str)
    return yaml.load(stream, UnicodeLoader)


def load_yaml_resource(package, filename):
    """Return the builtin configuration."""
    return yaml.safe_load(resource_string(package, filename))



# encoding: utf-8
"""
Python API for talking to the HIT online control/DVM.
"""

from __future__ import unicode_literals


__title__ = 'hit_csys'
__summary__ = 'Online control for the HIT accelerator facility.'
__uri__ = 'https://bitbucket.org/coldfix/hit-online-control'

__version__ = '0.7.0'

__author__ = 'Thomas Gläßle'
__email__ = 't_glaessle@gmx.de'

__license__ = None
__copyright__ = '(C) 2013 - 2018 HIT Betriebs GmbH'

# Trove classifiers: https://pypi.python.org/pypi?:action=list_classifiers
__classifiers__ = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Healthcare Industry',
    'Intended Audience :: Science/Research',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering :: Medical Science Apps.',
    'Topic :: Scientific/Engineering :: Physics',
]

entry_points = """
[madgui.online.PluginLoader]
stub = hit_csys.plugin:StubLoader
dll = hit_csys.plugin:DllLoader
"""



"""
Start an interactive testing console for the BeamOptikDLL wrapper.

Note that this cannot equivalently be done without the GUI: The
'BeamOptikDLL.dll' library creates an application window and therefore
requires a main loop to process window messages. This module uses wxPython_
to create the main loop and PyCrust_ to provide an interactive python
console.

.. _wxPython: http://wxpython.org/
.. _PyCrust: http://wxpython.org/py.php

The DLL is connected on startup and the wrapper object is stored in the
global variable ``dvm``.
"""

from __future__ import absolute_import

import wx
import wx.py.crust
import logging

from .beamoptikdll import BeamOptikDLL


class App(wx.App):

    def OnInit(self):
        loc = {}
        frame = wx.py.crust.CrustFrame(locals=loc)
        frame.Show()
        frame.crust.shell.redirectStdout()
        frame.crust.shell.redirectStderr()
        logging.basicConfig(level=logging.INFO)
        loc['frame'] = frame
        loc['dvm'] = BeamOptikDLL.load_library()
        loc['dvm'].GetInterfaceInstance()
        return True


def main():
    """Invoke GUI application."""
    app = App(redirect=False)
    app.MainLoop()


if __name__ == '__main__':
    main()



"""
Python wrapper for 'BeamOptikDLL.dll'.
"""

# NOTE: This module deliberately depends only on the standard library and can
# therefore be simply included in another application without having to
# install or provide anything else (except for the actual DLL of course).

from collections import namedtuple
from ctypes import c_double as Double, c_int as Int
import ctypes
import logging
import platform


def _encode(s):
    return s if isinstance(s, bytes) else s.encode('utf-8')

def _decode(s):
    return s.decode('utf-8') if isinstance(s, bytes) else s


if platform.architecture()[0] == '64bit':
    _Str = ctypes.c_wchar_p         # constructor wants unicode
    def Str(s):
        return _Str(_decode(s))
else:
    _Str = ctypes.c_char_p          # constructor wants bytes
    def Str(s):
        return _Str(_encode(s))


EFI = namedtuple('EFI', ['energy', 'focus', 'intensity', 'gantry_angle'])


class _EnumBase(int):

    """Abstract base type for enums (missed :cvar:`_value_names`)."""

    def __repr__(self):
        return '<{}.{}: {}>'.format(
            self.__class__.__name__,
            self._value_names[int(self)],
            int(self),
        )


def make_enum(name, value_names):
    """Create a simple enum type (like in C++)."""
    Enum = type(name, (_EnumBase,), {
        '_value_names': value_names,
    })
    for i, v in enumerate(value_names):
        setattr(Enum, v, Enum(i))
    return Enum


DVMStatus = make_enum('DVMStatus', [
    'Stop', 'Idle', 'Init', 'Ready', 'Busy', 'Finish', 'Error'
])
GetOptions = make_enum('GetOptions', [
    'Current', 'Saved'
])
ExecOptions = make_enum('ExecOptions', [
    'CalcAll', 'CalcDif', 'SimplyStore'
])
GetSDOptions = make_enum('GetSDOptions', [
    'Current', 'Database', 'Test'
])


class BeamOptikDLL(object):

    """
    Thin wrapper around the BeamOptikDLL API.

    It abstracts the ctypes data types and automates InterfaceId as well as
    iDone. Nothing else.

    Instanciation is a two-step process as follows:

    >>> obj = BeamOptikDLL.load_library()
    >>> obj.GetInterfaceInstance()
    """

    if platform.architecture()[0] == '64bit':
        filename = 'BeamOptikDLL64.dll'
    else:
        filename = 'BeamOptikDLL.dll'

    #----------------------------------------
    # internal methods
    #----------------------------------------

    error_messages = [
        None,
        "Invalid Interface ID.",
        "Parameter not found in internal DVM list.",
        "GetValue failed.",
        "SetValue failed.",
        "Unknown option.",
        "Memory error.",
        "General runtime error.",
        "Ramp event not supported.",
        "Ramp data not available.",
        "Invalid offset for ramp function."]

    @classmethod
    def check_return(cls, done):
        """
        Check DLL-API exit code for errors and raise exception.

        :param int done: exit code of an DLL function
        :raises RuntimeError: if the exit code is a known error code != 0
        :raises ValueError: if the exit code is unknown
        """
        if 0 < done and done < len(cls.error_messages):
            raise RuntimeError(cls.error_messages[done])
        elif done != 0:
            raise ValueError("Unknown error: %i" % done)

    def _call(self, function, *params):
        """
        Call the specified DLL function.

        :param str function: name of the function to call
        :param params: ctype function parameters except for piDone.
        :raises RuntimeError: if the exit code indicates any error

        For internal use only!
        """
        done = Int()
        params = list(params)
        if function == 'SelectMEFI':
            params.insert(6, done)
        else:
            params.append(done)
        def param(p):
            return p if isinstance(p, _Str) else ctypes.byref(p)
        func = getattr(self.lib, function)
        args = tuple(map(param, params))
        func(*args)
        self.check_return(done.value)

    #----------------------------------------
    # things that don't require IID to be set:
    #----------------------------------------

    @classmethod
    def load_library(cls, filename=filename):
        """
        Search for the DLL in PATH and return a BeamOptikDLL wrapper object.

        NOTE: for the 64bit library, you also need 'ParamDownloads64.dll' in
        your PATH.
        """
        try:
            return cls(ctypes.windll.LoadLibrary(filename))
        except AttributeError:
            raise OSError("BeamOptikDLL.dll only available on windows.")

    @classmethod
    def check_library(cls):
        """Check if library is available."""
        try:
            ctypes.windll
        except AttributeError:
            # Not available for Linux:
            return False
        # TODO: try to find DLL, return False if not in PATH
        return True

    def DisableMessageBoxes(self):
        """
        Prevent creation of certain message boxes.

        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('DisableMessageBoxes')

    def GetInterfaceInstance(self):
        """
        Initialize a BeamOptikDLL instance (connects DB and initialize DLL).

        :return: new instance id
        :rtype: int
        :raises RuntimeError: if the exit code indicates any error
        """
        if self._iid is not None:
            raise RuntimeError("GetInterfaceInstance cannot be called twice.")
        iid = Int()
        self._call('GetInterfaceInstance', iid)
        self._iid = iid
        return iid

    @property
    def lib(self):
        """Shared library proxy."""
        return self._lib

    #----------------------------------------
    # object API
    #----------------------------------------

    def __init__(self, lib):
        """
        Initialize member variables.

        Usually, you want to use :classmethod:`load_library` to create
        instances instead of directly invoking this constructor.

        :param lib: shared library proxy object
        """
        self._lib = lib
        self._iid = None
        self._selected_vacc = None
        self._selected_efi = EFI(None, None, None, None)
        self._logger = logging.getLogger(__name__)

    @property
    def iid(self):
        """Interface instance ID."""
        if self._iid is None:
            raise RuntimeError("GetInterfaceInstance must be called before using other methods.")
        return self._iid

    def __bool__(self):
        """Check if the object belongs to an initialized interface instance."""
        return self._iid is not None

    __nonzero__ = __bool__

    def FreeInterfaceInstance(self):
        """
        Free resources.

        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('FreeInterfaceInstance', self.iid)
        self._iid = None

    def GetDVMStatus(self):
        """
        Get current status of selected virtual accelerator.

        :return: DVM status
        :rtype: DVMStatus
        :raises RuntimeError: if the exit code indicates any error
        """
        status = Int()
        self._call('GetDVMStatus', self.iid, status)
        return DVMStatus(status.value)

    def SelectVAcc(self, vaccnum):
        """
        Select the virtual accelerator.

        :param int vaccnum: virtual accelerator number (0-255)
        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('SelectVAcc', self.iid, Int(vaccnum))
        self._selected_vacc = vaccnum

    def SelectMEFI(self, vaccnum, energy, focus, intensity, gantry_angle=0):
        """
        Select EFI combination for the currently selected VAcc.

        :param int vaccnum: virtual accelerator number (0-255)
        :param int energy: energy channel (1-255)
        :param int focus: focus channel (1-6)
        :param int intensity: intensity channel (1-15)
        :param int gantry_angle: gantry angle index (1-36)
        :return: physical EFI values
        :rtype: EFI
        :raises RuntimeError: if the exit code indicates any error

        CAUTION: SelectVAcc must be called before invoking this function!
        """
        values = [Double(), Double(), Double(), Double()]
        self._call('SelectMEFI', self.iid, Int(vaccnum),
                   Int(energy), Int(focus), Int(intensity), Int(gantry_angle),
                   *values)
        if vaccnum == self._selected_vacc:
            self._selected_efi = EFI(energy, focus, intensity, gantry_angle)
        else:
            self._logger.warn('You must call SelectVAcc() before SelectMEFI()!')
        return EFI(*[v.value for v in values])

    def GetSelectedVAcc(self):
        """
        Get selected virtual accelerator.

        :return: virtual accelerator number (0-255)
        :rtype: int
        :raises RuntimeError: if the exit code indicates any error
        """
        vaccnum = Int()
        self._call('GetSelectedVAcc', self.iid, vaccnum)
        return vaccnum.value

    def GetFloatValue(self, name, options=GetOptions.Current):
        """
        Get parameter value.

        :param str name: parameter name
        :param GetOptions options: options
        :return: parameter value
        :rtype: float
        :raises RuntimeError: if the exit code indicates any error
        """
        value = Double()
        self._call('GetFloatValue', self.iid, Str(name), value, Int(options))
        return value.value

    def SetFloatValue(self, name, value, options=0):
        """
        Set parameter value.

        :param str name: parameter name
        :param float value: parameter value
        :param options: not used currently
        :raises RuntimeError: if the exit code indicates any error

        Changes take effect after calling :func:`ExecuteChanges`.
        """
        self._call('SetFloatValue', self.iid, Str(name), Double(value), Int(options))

    def ExecuteChanges(self, options):
        """
        Apply parameter changes.

        :param ExecOptions options: what to do exactly
        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('ExecuteChanges', self.iid, Int(options))

    def SetNewValueCallback(self, callback):
        """
        Install a callback function that should be called whenever the
        calculation of a new value is complete (…whatever that means).

        :param callback: ``callable(name:str, val:float, type:int)``
        :raises RuntimeError: if the exit code indicates any error
        """
        PTR = ctypes.POINTER
        @ctypes.WINFUNCTYPE(None, PTR(_Str), PTR(Double), PTR(Int))
        def c_callback(name, value, type_):
            return callback(_decode(name.value),
                            value.contents.value,
                            type_.contents.value)
        self._call('SetNewValueCallback', self.iid, c_callback)

    def GetFloatValueSD(self, name, options=0):
        """
        Get current beam measurement at specific element.

        :param str name: parameter name (<observable>_<element name>)
        :param GetSDOptions options: options
        :return: measured value
        :rtype: float
        :raises RuntimeError: if the exit code indicates any error
        """
        value = Double()
        self._call('GetFloatValueSD', self.iid, Str(name), value, Int(options))
        return value.value

    def GetLastFloatValueSD(self, name, vaccnum,
                            energy, focus, intensity, gantry_angle=0,
                            options=0):
        """
        Get previous beam measurement at specific element.

        :param str name: parameter name (<observable>_<element name>)
        :param int vaccnum: virtual accelerator number (0-255)
        :param GetSDOptions options: options
        :return: measured value and EFI combination
        :rtype: tuple
        :raises RuntimeError: if the exit code indicates any error
        """
        value = Double()
        self._call('GetLastFloatValueSD', self.iid, Str(name),
                   value, Int(vaccnum), Int(options),
                   Int(energy), Int(focus), Int(intensity), Int(gantry_angle))
        return value.value

    def StartRampDataGeneration(self, vaccnum, energy, focus, intensity):
        """
        Call StartRampDataGeneration().
        """
        order_num = Int()
        self._call('StartRampDataGeneration', self.iid,
                   Int(vaccnum), Int(energy), Int(focus), Int(intensity),
                   order_num)
        sel_efi = self._selected_efi
        if (vaccnum != self._selected_vacc or
            energy != sel_efi.energy or
            focus != sel_efi.focus or
            intensity != sel_efi.intensity):
            self._logger.warn("You must call SelectEFI() before StartRampDataGeneration()!")
        return order_num.value

    def GetRampDataValue(self, order_num, event_num, delay,
                         parameter_name, device_name):
        """
        Call GetRampDataValue()
        """
        value = Double()
        self._call('GetRampDataValue', self.iid,
                   Int(order_num), Int(event_num), Int(delay),
                   Str(parameter_name), Str(device_name),
                   value)
        return value.value

    def SetIPC_DVM_ID(self, name):
        """Call SetIPC_DVM_ID(). Not implemented!"""
        raise NotImplementedError() # TODO

    def GetMEFIValue(self):
        """
        Retrieve EFI values for current selection.

        :return: physical EFI values, EFI channel numbers
        :rtype: tuple(EFI, EFI)
        :raises RuntimeError: if the exit code indicates any error
        """
        values = [Double(), Double(), Double(), Double()]
        channels = [Int(), Int(), Int(), Int()]
        self._call('GetMEFIValue', self.iid, *(values + channels))
        return (EFI(*[v.value for v in values]),
                EFI(*[c.value for c in channels]))



"""
Stub class for BeamOptikDLL.dll ctypes proxy objects as used by
:class:`~hit_csys.beamoptikdll.BeamOptikDLL`. Primarily used for
offline testing of the basic functionality.
"""

import functools
import ctypes

from madgui.core.unit import madx_units

from . import beamoptikdll

import random

c_str = ctypes.c_char_p, ctypes.c_wchar_p


__all__ = [
    'BeamOptikDllProxy',
]


def _unbox(param):
    """Unbox a call parameter created by ctypes.byref."""
    return param.value if isinstance(param, c_str) else param._obj


def _api_meth(func):

    """
    Decorator for methods conforming to the BeamOptikDLL API.

    Unboxes parameter references and sets the ``done`` from the function's
    return value.
    """

    @functools.wraps(func)
    def wrapper(self, *args):
        idone = 6 if func.__name__ == 'SelectMEFI' else len(args) - 1
        done = _unbox(args[idone])
        args = args[:idone] + args[idone+1:]
        done.value = 0
        unboxed_args = tuple(map(_unbox, args))
        if self.logger:
            self.logger.info('{}{}'.format(func.__name__, unboxed_args))
        ret = func(self, *unboxed_args)
        if ret is not None:
            done.value = ret

    return wrapper


class BeamOptikDllProxy(object):

    """A fake implementation for a ctypes proxy of the BeamOptikDLL."""

    # TODO: Support read-only/write-only parameters
    # TODO: Prevent writing unknown parameters by default

    def __init__(self, frame, logger=None):
        """Initialize new library instance with no interface instances."""
        self.params = {}
        self.instances = {}
        self.logger = logger
        self.next_iid = 0
        self.frame = frame
        self.jitter = True

    def on_model_changed(self):
        self.model = self.frame.model
        self.params.clear()
        self.frame.control.write_all()
        if self.jitter:
            for k in self.params:
                self.params[k] *= random.uniform(0.95, 1.1)

    @_api_meth
    def DisableMessageBoxes(self):
        """Do nothing. There are no message boxes anyway."""
        pass

    @_api_meth
    def GetInterfaceInstance(self, iid):
        """Create a new interface instance."""
        iid.value = self.next_iid
        self.instances[iid.value] = {
            'VAcc': 1,
            'EFIA': (1, 1, 1, 1),
        }
        self.next_iid += 1

    @_api_meth
    def FreeInterfaceInstance(self, iid):
        """Destroy a previously created interface instance."""
        assert self.instances[iid.value]
        self.instances[iid.value] = None

    @_api_meth
    def GetDVMStatus(self, iid, status):
        """Get DVM ready status."""
        assert self.instances[iid.value]
        # The test lib has no advanced status right now.
        status.value = beamoptikdll.DVMStatus.Ready

    @_api_meth
    def SelectVAcc(self, iid, vaccnum):
        """Set virtual accelerator number."""
        assert self.instances[iid.value]
        self.instances[iid.value]['VAcc'] = vaccnum.value

    @_api_meth
    def SelectMEFI(self, iid, vaccnum,
                   energy, focus, intensity, gantry_angle,
                   energy_val, focus_val, intensity_val, gantry_angle_val):
        """Set MEFI in current VAcc."""
        # The real DLL requires SelectVAcc to be called in advance, so we
        # enforce this constraint here as well:
        assert self.instances[iid.value]['VAcc'] == vaccnum.value
        self.instances[iid.value]['EFIA'] = (
            energy.value,
            focus.value,
            intensity.value,
            gantry_angle.value,
        )
        energy_val.value = float(energy.value)
        focus_val.value = float(focus.value)
        intensity_val.value = float(intensity.value)
        gantry_angle_val.value = float(gantry_angle.value)

    @_api_meth
    def GetSelectedVAcc(self, iid, vaccnum):
        """Get currently selected VAcc."""
        vaccnum.value = self.instances[iid.value]['VAcc']

    @_api_meth
    def GetFloatValue(self, iid, name, value, options):
        """Get a float value from the "database"."""
        assert self.instances[iid.value]
        value.value = float(self.params.get(name, 0))

    @_api_meth
    def SetFloatValue(self, iid, name, value, options):
        """Store a float value to the "database"."""
        assert self.instances[iid.value]
        self.params[name] = value.value

    @_api_meth
    def ExecuteChanges(self, iid, options):
        """Do nothing: our "database" is currently non-transactional."""
        assert self.instances[iid.value]

    @_api_meth
    def SetNewValueCallback(self, iid, callback):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def GetFloatValueSD(self, iid, name, value, options):
        """Get beam diagnostic value."""
        assert self.instances[iid.value]

        par_name, el_name = name.lower().split('_', 1)
        index = self.model.elements.index(el_name)
        index = self.model.indices[index].stop

        cols = {
            'widthx': 'envx',
            'widthy': 'envy',
            'posx': 'posx',
            'posy': 'posy',
        }
        col = cols[par_name]
        twiss = self.model.get_twiss_column(col)
        v = madx_units.strip_unit(col, twiss[index])

        if self.jitter:
            if par_name in ('widthx', 'widthy'):
                v *= random.uniform(0.95, 1.1)
            elif par_name in ('posx', 'posy'):
                v += random.uniform(-0.0005, 0.0005)

        value.value = v * 1000

    @_api_meth
    def GetLastFloatValueSD(self, iid,
                            name, value, vaccnum, options,
                            energy, focus, intensity, gantry_angle):
        """Get beam diagnostic value."""
        # behave exactly like GetFloatValueSD and ignore further parameters
        # for now
        self.GetFloatValueSD.__wrapped__(self, iid, name, value, options)

    @_api_meth
    def StartRampDataGeneration(self, iid,
                                vaccnum, energy, focus, intensity, order_num):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def GetRampDataValue(self, iid, order_num, event_num, delay,
                         parameter_name, device_name, value):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def SetIPC_DVM_ID(self, iid, name):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def GetMEFIValue(self, iid,
                     energy_val, focus_val, intensity_val, gantry_angle_val,
                     energy_chn, focus_chn, intensity_chn, gantry_angle_chn):
        """Get current MEFI combination."""
        e, f, i, a = self.instances[iid.value]['EFIA']
        energy_chn.value = e
        focus_chn.value = f
        intensity_chn.value = i
        gantry_angle_chn.value = a
        energy_val.value = float(energy_chn.value)
        focus_val.value = float(focus_chn.value)
        intensity_val.value = float(intensity_chn.value)
        gantry_angle_val.value = float(gantry_angle_chn.value)



# encoding: utf-8
"""
Tools to work with DVM paramater list.
"""

from collections import namedtuple

from madgui.core import unit
from hit_csys.util import csv_unicode_reader


DVM_Parameter = namedtuple('DVM_Parameter', [
    'name',
    'ui_name',
    'ui_hint',
    'ui_prec',
    'unit',
    'ui_unit',
    'ui_conv',
])


#----------------------------------------
# CSV column types
#----------------------------------------

def CsvStr(s):
    return s


def CsvInt(s):
    return int(s) if s else None


def CsvFloat(s):
    return float(s) if s else None


def CsvUnit(s):
    if s == '%':
        return 0.01
    s = s.replace(u'grad', u'degree')
    s = s.replace(u'Ohm', u'ohm')
    s = s.replace(u'part.', u'count')   # used for particle count
    return unit.from_config(s)


def load_csv(lines, encoding='utf-8', delimiter=';'):
    """
    Parse DVM parameters from CSV file exported from XLS documentation
    spreadsheet (e.g. DVM-Parameter_v2.10.0-10-HIT.xls)
    """
    return load_csv_data(csv_unicode_reader(
        lines, encoding=encoding, delimiter=delimiter))


def load_csv_data(rows):
    return _parse_csv_data(rows)


def _parse_csv_data(rows):
    parse_row = lambda row: DVM_Parameter(**{
        n: _csv_column_types[n](row[i].strip())
        for n, i in _csv_column_index.items()
    })
    cluster_name = ''
    cluster_items = []
    for row in rows:
        item = parse_row(row)
        # detect cluster header lines:
        link = row[0]
        if link and not link.isdigit() and not item.name:
            # yield previous element/context
            if cluster_items:
                yield (cluster_name, cluster_items)
            cluster_name = link
            cluster_items = []
        elif item.name:
            cluster_items.append(item)
    if cluster_items:
        yield (cluster_name, cluster_items)

# all columns in csv file:
_csv_column_names = [
    '',                 # Nr. für Link
    'name',             # Code Param (GSI-Nomenklatur)
    '',                 # Code Gerät (GSI- NomenkLatur) entspr. DCU!
    '',                 # Code Gruppe (=Kalkulationsgruppe); möglichst GSI-NomenkLatur
    'ui_name',          # GUI Beschriftung Parameter (ohne Einheit)
    'ui_hint',          # GUI Beschriftung Hint
    '',                 # Position ExpertGrids
    '',                 # DVM liest Parameter
    '',                 # DVM ändert Parameter
    '',                 # DVM Datensatz spezifisch
    '',                 # Rein temporär
    '',                 # MEFI-Abhängigkeit
    '',                 # Input Param wird Output Param bei MEFI
    '',                 # In Gui Init änderbar
    '',                 # Daten-typ
    'ui_prec',          # Präzision (Anz. Nachkomma im GUI)
    'unit',             # Einheit Parameter
    'ui_unit',          # Einheit Anzeige im GUI
    'ui_conv',          # Umrechnungsfaktor Einheit--> Einheit GUI
    '',                 # Beispielwert für Test in Einheit GUI
    '',                 # Referenz auf DCU /MDE
    '',                 # (nicht verwendet)
    '',                 # Zugriffscode / editierbarkeit
    '',                 # Versions-  Relevanz
    '',                 # Detail Ansicht verfügbar (ja/nein)
    '',                 # Link auf Maximalwert
    '',                 # Link auf Minimalwert
    '',                 # Code Min/Max- Rechen-vorschrift
    '',                 # Master-gruppe
    '',                 # Defaultwert Änderung pro Pfeiltasten-druck/Maus-radsegment in Einheit GUI
    '',                 # Im laufenden Betrieb änderbar (ja/ nein)
    '',                 # Link auf zugehörigen sekundären Wert
]

_csv_column_types = {
    'name':     CsvStr,
    'ui_name':  CsvStr,
    'ui_hint':  CsvStr,
    'ui_prec':  CsvInt,
    'unit':     CsvUnit,
    'ui_unit':  CsvUnit,
    'ui_conv':  CsvFloat,
}

# inverse map of _csv_columns[i] (used columns)
_csv_column_index = {
    name: index
    for index, name in enumerate(_csv_column_names)
    if name
}



# encoding: utf-8
"""
Madgui online control plugin.
"""

from __future__ import absolute_import

import logging
import itertools
try:
    from importlib.resources import resource_stream     # faster import
except ImportError:
    from pkg_resources import resource_stream

from pydicti import dicti

from .util import load_yaml_resource

from .beamoptikdll import BeamOptikDLL, ExecOptions
from .stub import BeamOptikDllProxy

from madgui.core.base import Object, Signal
from madgui.qt import QtGui
from madgui.core import unit
from madgui.online import api

from .dvm_parameters import load_csv, DVM_Parameter


class StubLoader(api.PluginLoader):

    title = '&test stub'
    descr = 'a stub version (for offline testing)'
    hotkey = 'Ctrl+C'

    @classmethod
    def check_avail(cls):
        return True

    @classmethod
    def load(cls, frame):
        # logger = frame.getLogger('hit_csys.stub')
        logger = logging.getLogger('hit_csys.stub')
        proxy = BeamOptikDllProxy(frame, logger)
        dvm = BeamOptikDLL(proxy)
        dvm.on_model_changed = proxy.on_model_changed
        params = load_dvm_parameters()
        return HitOnlineControl(dvm, params, frame)


class DllLoader(api.PluginLoader):

    title = '&online control'
    descr = 'the online control'
    hotkey = None

    @classmethod
    def check_avail(cls):
        return BeamOptikDLL.check_library()

    @classmethod
    def load(cls, frame):
        """Connect to online database."""
        dvm = BeamOptikDLL.load_library()
        params = load_dvm_parameters()
        return HitOnlineControl(dvm, params, frame)


def load_dvm_parameters():
    with resource_stream('hit_csys', 'DVM-Parameter_v2.10.0-HIT.csv') as f:
        parlist = load_csv(f, 'utf-8')
    return dicti(
        (el_name, dicti((p.name, p) for p in params))
        for el_name, params in parlist)


def _get_sd_value(dvm, el_name, param_name):
    """Return a single SD value (with unit)."""
    sd_name = param_name + '_' + el_name
    plain_value = dvm.GetFloatValueSD(sd_name.upper())
    return plain_value / 1000       # mm to m


class HitOnlineControl(api.OnlinePlugin):

    def __init__(self, dvm, params, frame):
        self._dvm = dvm
        self._params = params
        self._frame = frame
        self._config = load_yaml_resource('hit_csys', 'config.yml')

    # OnlinePlugin API

    def connect(self):
        """Connect to online database (must be loaded)."""
        self._dvm.GetInterfaceInstance()
        self._frame.model_changed.connect(self.on_model_changed)
        self.on_model_changed()

    def disconnect(self):
        """Disconnect from online database."""
        self._dvm.FreeInterfaceInstance()
        self._frame.model_changed.disconnect(self.on_model_changed)

    def on_model_changed(self):
        if hasattr(self._dvm, 'on_model_changed'):
            self._dvm.on_model_changed()

    @property
    def _model(self):
        return self._frame.model

    def execute(self, options=ExecOptions.CalcDif):
        """Execute changes (commits prior set_value operations)."""
        self._dvm.ExecuteChanges(options)

    def param_info(self, knob):
        """Get parameter info for backend key."""
        if isinstance(knob, Knob):
            return knob.info
        el_name = knob.split('_', 1)[1]
        try:
            return self._params[el_name][knob]
        except KeyError:
            return None

    def read_monitor(self, name):
        """
        Read out one monitor, return values as dict with keys:

            widthx:     Beam x width
            widthy:     Beam y width
            posx:       Beam x position
            posy:       Beam y position
        """
        keys_backend = ('posx', 'posy', 'widthx', 'widthy')
        keys_internal = ('posx', 'posy', 'envx', 'envy')
        values = {}
        for src, dst in zip(keys_backend, keys_internal):
            # TODO: Handle usability of parameters individually
            try:
                val = _get_sd_value(self._dvm, name, src)
            except RuntimeError:
                return {}
            # TODO: move sanity check to later, so values will simply be
            # unchecked/grayed out, instead of removed completely
            # The magic number -9999.0 signals corrupt values.
            # FIXME: Sometimes width=0 is returned. ~ Meaning?
            if src.startswith('width') and val <= 0:
                return {}
            values[dst] = val
        return values

    def get_knob(self, elem, attr):
        """Return a :class:`Knob` belonging to the given attribute."""
        attr = attr.lower()
        el_name = elem['name'].lower()
        el_type = elem['type'].lower()
        if '_' in el_name:
            body, suffix = el_name.rsplit('_', 1)
            if suffix == 'corr' and attr == 'kick':
                el_name = body
        # Sometimes the parameter name does not include the element H/V
        # suffix, e.g.: R1MS1H -> ax_R1MS1, R1MS1V -> ay_R1MS1
        if el_name[-1:] in 'hv':
            suffixes = [el_name, el_name[:-1]]
        else:
            suffixes = [el_name]
        el_pars = self._params.get(el_name, {})
        el_expr = getattr(elem[attr], '_expression', '').lower()
        prefixes = PREFIXES.get((el_type, attr), [])
        # Prefer the parameter name defined by the expression; otherwise use
        # the element name plus known prefixes, such as 'ax_', 'ay_':
        par_names = [el_expr] + [
            prefix + '_' + suffix
            for prefix, suffix in itertools.product(prefixes, suffixes)
        ]
        for param in map(el_pars.get, par_names):
            if param:
                return Knob(self, elem, attr, param)
        if  (el_name.startswith('gant') and
             el_type == 'srotation' and
             attr =='angle'):
            param = DVM_Parameter(
                name='gantry_angle',
                ui_name='gantry_angle',
                ui_hint='',
                ui_prec=3,
                unit=1*unit.units.degree,
                ui_unit=1*unit.units.degree,
                ui_conv=1,
            )
            return MEFI_Param(self, elem, 'gantry', param, 3)

    def read_param(self, param):
        """Read parameter. Return numeric value."""
        return self._dvm.GetFloatValue(param)

    def write_param(self, param, value):
        """Update parameter into control system."""
        self._dvm.SetFloatValue(param, value)

    def get_beam(self):
        units  = unit.units
        e_para = ENERGY_PARAM.get(self._model.seq_name, 'E_HEBT')
        z_num  = self._dvm.GetFloatValue('Z_POSTSTRIP')
        mass   = self._dvm.GetFloatValue('A_POSTSTRIP') * units.u
        charge = self._dvm.GetFloatValue('Q_POSTSTRIP') * units.e
        e_kin  = (self._dvm.GetFloatValue(e_para) or 1) * units.MeV / units.u
        return {
            'particle': PERIODIC_TABLE[round(z_num)],
            'charge':   unit.from_ui('charge', charge),
            'mass':     unit.from_ui('mass',   mass),
            'energy':   unit.from_ui('energy', mass * (e_kin + 1*units.c**2)),
        }

ENERGY_PARAM = {
    'lebt': 'E_SOURCE',
    'mebt': 'E_MEBT',
}

PERIODIC_TABLE = {
    1: 'p',
    2: 'He',
    6: 'C',
    8: 'O',
}

# NOTE: order is important, so keep 'dax' before 'ax', etc:
PREFIXES = {
    ('sbend',   'angle'):  ['ax'],
    ('quadrupole', 'k1'):  ['kl_efg', 'kl'],
    ('quadrupole', 'k1s'):  ['kl_efg', 'kl'],
    ('hkicker',  'kick'):  ['dax', 'ax'],
    ('vkicker',  'kick'):  ['day', 'ay'],
    ('solenoid',   'ks'):  ['ks'],
    ('multipole', 'knl[0]'):  ['dax', 'ax'],
    ('multipole', 'ksl[0]'):  ['day', 'ay'],
}

CSYS_ATTR = { 'k1': 'kl', 'k1s': 'kl' }


class Knob(api.Knob):

    def __init__(self, plug, elem, attr, param):
        super().__init__(plug, elem, CSYS_ATTR.get(attr, attr),
                         param.name, param.unit)
        self.info = param


class MEFI_Param(Knob):

    def __init__(self, plug, elem, attr, param, idx):
        super().__init__(plug, elem, attr, param)
        self.idx = idx

    def read(self):
        return unit.madx_units.strip_unit(
            self.attr, unit.add_unit(
                self.plug._dvm.GetMEFIValue()[0][self.idx],
                self.info.unit))

    def write(self, value):
        pass
        #raise NotImplementedError(
        #    "Must change MEFI parameters via BeamOptikDLL GUI")



"""
Start an interactive testing console for the BeamOptikDLL wrapper.

Note that this cannot equivalently be done without the GUI: The
'BeamOptikDLL.dll' library creates an application window and therefore
requires a main loop to process window messages. This module uses PyQt4_
to create the main loop and spyderlib_ to provide an interactive python
console.

.. _PyQt4: https://riverbankcomputing.com/software/pyqt/intro
.. _spyderlib: https://github.com/spyder-ide/spyder

The DLL is connected on startup and the wrapper object is stored in the
global variable ``dvm``.
"""

from __future__ import absolute_import

import sys
import signal
import logging

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.qt import QtCore, QtGui

from .beamoptikdll import BeamOptikDLL
from .stub import BeamOptikDllProxy


def create(user_ns):
    """Create an in-process kernel."""
    manager = QtInProcessKernelManager()
    manager.start_kernel(show_banner=False)
    kernel = manager.kernel
    kernel.gui = 'qt4'
    kernel.user_ns = user_ns

    client = manager.client()
    client.start_channels()

    widget = RichJupyterWidget()
    widget.kernel_manager = manager
    widget.kernel_client = client
    return widget


class MainWindow(QtGui.QMainWindow):

    def __init__(self, namespace):
        QtGui.QMainWindow.__init__(self)
        self.ns = namespace
        self.shell = create(namespace)
        # self.shell.interpreter.restore_stds()
        self.setCentralWidget(self.shell)
        QtCore.QTimer.singleShot(0, self.load_dll)

    def closeEvent(self, event):
        event.accept()

    def load_dll(self):
        ns = self.ns
        ns['window'] = self
        ns['dvm'] = BeamOptikDLL.load_library()
        # proxy = BeamOptikDllProxy({})
        # ns['dvm'] = BeamOptikDLL(proxy)
        ns['dvm'].GetInterfaceInstance()


def main():
    """Invoke GUI application."""
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = QtGui.QApplication(sys.argv)
    ns = {}
    ns['exit'] = sys.exit
    window = MainWindow(ns)

    logging.basicConfig(level=logging.INFO)

    window.show()
    app.exec_()


if __name__ == '__main__':
    main()



"""
Loads BeamOptikDLL in a python process and waits in a simple win32 message
loop. This module does not provide any custom/interactive GUI apart from the
GUI provided by BeamOptikDLL itself.
"""

from __future__ import absolute_import

import logging
import platform
import win32gui

from .beamoptikdll import BeamOptikDLL


def main():
    """Invoke GUI application."""
    logging.basicConfig(level=logging.INFO)
    dll = BeamOptikDLL.load_library()
    dll.GetInterfaceInstance()
    win32gui.PumpMessages()


if __name__ == '__main__':
    main()



# encoding: utf-8
from setuptools import setup
from distutils.util import convert_path


def read_file(path):
    """Read a file in binary mode."""
    with open(convert_path(path), 'rb') as f:
        return f.read()


def exec_file(path):
    """Execute a python file and return the `globals` dictionary."""
    namespace = {}
    exec(read_file(path), namespace, namespace)
    return namespace


def main():
    """Execute setup."""
    long_description = read_file('README.rst').decode('utf-8')
    meta = exec_file('hit_csys/__init__.py')
    setup(
        name='hit_csys',
        version=meta['__version__'],
        description=meta['__summary__'],
        long_description=long_description,
        author=meta['__author__'],
        author_email=meta['__email__'],
        url=meta['__uri__'],
        license=meta['__license__'],
        classifiers=meta['__classifiers__'],
        packages=[
            'hit_csys',
        ],
        entry_points={
            'gui_scripts': [
                'hit_csys = hit_csys.gui_wx:main'
            ],
            'madgui.online.PluginLoader': [
                'stub = hit_csys.plugin:StubLoader',
                'dll = hit_csys.plugin:DllLoader',
            ]
        },
        install_requires=[
            'pydicti>=0.0.4',
            'PyYAML',
        ],
        include_package_data=True,  # install files matched by MANIFEST.in
    )


if __name__ == '__main__':
    main()
